#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <signal.h>
#include <sys/mman.h>

#ifdef __XENO__

#include <posix/posix.h>
#include <rtdk.h>
#include <native/task.h>

#else /* !__XENO__ */

#include <pthread.h>
#include <semaphore.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <mqueue.h>
#include <errno.h>

#endif /* __XENO__ */

#define NUM_ARGS		2
#define TASK_STACK_SIZE		2048
#define TASK_PRIO		98
#define BREAD_SUM_LEN		(1024*1024)
#define NUM_ITER		10


#define DEBUG

#ifdef DEBUG

#define dprintf(fmt,args...)	printf(fmt,##args)
#define dperror(fmt)		perror(fmt)

#else

#define dprintf(fmt,args...)
#define dperror(fmt)

#endif


#define MASTER_OP(rd, wr, idx)					\
	do {							\
		pthread_mutex_unlock (wr);			\
		dprintf ("[thrd_%02d] rel_[%8p]\n",(idx),(wr));	\
		bread(); bread(); bread(); bread(); bread();	\
		pthread_yield();				\
		dprintf ("[thrd_%02d] try_[%8p]\n",(idx),(rd));	\
		pthread_mutex_lock (rd);			\
		dprintf ("[thrd_%02d] acq_[%8p]\n",(idx),(rd));	\
	} while (0)


typedef struct _task_state {
	int idx;
	int iter;
	pthread_mutex_t* rd_lock;
	pthread_mutex_t* wr_lock;
} state;



static long buf[BREAD_SUM_LEN];
static volatile float overhead, actual;

static struct timespec start, stop;

void start_time (void)	{
	start.tv_sec	= start.tv_nsec	= 0;
	stop.tv_sec	= stop.tv_nsec	= 0;
	clock_gettime (CLOCK_REALTIME, &start);
}

void stop_time (void)	{ 
	clock_gettime (CLOCK_REALTIME, &stop); 
}

float gettime (void)	{
	return ( ((stop.tv_sec  - start.tv_sec)  * 1000000.) +
		 ((stop.tv_nsec - start.tv_nsec) / 1000.) );

}



long bread(void)
{

	long sum = 0;
	long nbytes = BREAD_SUM_LEN;
	register long *p;
	register long *next;

	register long *end;

	p = (long*)buf;
	end = (long*)buf + nbytes;
	for (next = p + 128; (long*)next<=(long*)end; p = next, next += 128) {
		sum +=
			p[0]+p[1]+p[2]+p[3]+p[4]+p[5]+p[6]+p[7]+
			p[8]+p[9]+p[10]+p[11]+p[12]+p[13]+p[14]+
			p[15]+p[16]+p[17]+p[18]+p[19]+p[20]+p[21]+
			p[22]+p[23]+p[24]+p[25]+p[26]+p[27]+p[28]+
			p[29]+p[30]+p[31]+p[32]+p[33]+p[34]+p[35]+
			p[36]+p[37]+p[38]+p[39]+p[40]+p[41]+p[42]+
			p[43]+p[44]+p[45]+p[46]+p[47]+p[48]+p[49]+
			p[50]+p[51]+p[52]+p[53]+p[54]+p[55]+p[56]+
			p[57]+p[58]+p[59]+p[60]+p[61]+p[62]+p[63]+
			p[64]+p[65]+p[66]+p[67]+p[68]+p[69]+p[70]+
			p[71]+p[72]+p[73]+p[74]+p[75]+p[76]+p[77]+
			p[78]+p[79]+p[80]+p[81]+p[82]+p[83]+p[84]+
			p[85]+p[86]+p[87]+p[88]+p[89]+p[90]+p[91]+
			p[92]+p[93]+p[94]+p[95]+p[96]+p[97]+p[98]+
			p[99]+p[100]+p[101]+p[102]+p[103]+p[104]+
			p[105]+p[106]+p[107]+p[108]+p[109]+p[110]+
			p[111]+p[112]+p[113]+p[114]+p[115]+p[116]+
			p[117]+p[118]+p[119]+p[120]+p[121]+p[122]+
			p[123]+p[124]+p[125]+p[126]+p[127];
	}
	for (next = p + 16; (long*)next <= (long*)end; p = next, next += 16) {
		sum +=
			p[0]+p[1]+p[2]+p[3]+p[4]+p[5]+p[6]+p[7]+
			p[8]+p[9]+p[10]+p[11]+p[12]+p[13]+p[14]+
			p[15];
	}
	for (next = p + 1; (long*)next <= (long*)end; p = next, next++) {
		sum += *p;
	}
	return sum;
}


void* rt_task_slave (void*param)
{
	int idx		= ((state*)param)->idx;
	int iter	= ((state*)param)->iter;

	pthread_mutex_t* rd	= ((state*)param)->rd_lock;
	pthread_mutex_t* wr	= ((state*)param)->wr_lock;
	
	dprintf ("[thrd_%02d] rd: [%8p], wr: [%8p]\n",
			idx, rd, wr);

	for (; iter >= 0; iter--)
	{
		dprintf ("[thrd_%02d] try_[%8p]\n", idx, rd);
		pthread_mutex_lock (rd);
		dprintf ("[thrd_%02d] acq_[%8p]\n", idx, rd);
		bread(); bread(); bread();  bread(); bread(); 
		pthread_mutex_unlock (wr);
		dprintf ("[thrd_%02d] rel_[%8p]\n", idx, wr);
		pthread_yield();
	}

	dprintf ("[%s] exiting thread_num<%02d> \n",
		       	__FUNCTION__,idx);
	return NULL;
}

void* rt_task_master (void* param)
{
	int iter	= 0;
	float warmup	= 0;

	int cnt		= ((state*)param)->iter;
	int idx		= ((state*)param)->idx;

	pthread_mutex_t* rd	= ((state*)param)->rd_lock;
	pthread_mutex_t* wr	= ((state*)param)->wr_lock;

	dprintf ("[thrd_%02d] rd: [%8p], wr: [%8p]\n",
			idx, rd, wr);

	/* Warmup  */
	dprintf ("[thrd_%02d] *********** Warmup ***********\n", idx);
	start_time();
	for (iter = cnt; iter; iter--) 	MASTER_OP (rd, rd, idx);
	stop_time();	
	warmup = gettime();

	/* Dry run  */
	dprintf ("[thrd_%02d] *********** Dry run ***********\n", idx);
	start_time();
	for (iter = cnt; iter; iter--) 	MASTER_OP (rd, rd, idx);
	stop_time();	
	overhead = gettime();

	/* Full on: Loop the token around */
	dprintf ("[thrd_%02d] *********** Full-on run ***********\n", idx);
	start_time();
	for (iter = cnt; iter>=0 ; iter--) 	MASTER_OP (rd, wr, idx);
	stop_time();	
	actual = gettime();

	dprintf ("[%s] exiting thread_num<%02d> \n",
		       	__FUNCTION__,idx);
	return NULL;
}

void destroy_mutexes (pthread_mutex_t* mutex_list, int cnt)
{
	int iter = 0;

	for (iter = 0; iter <= cnt; iter++)
	{
		pthread_mutex_unlock ( &(mutex_list[iter]) );
		pthread_mutex_destroy ( &(mutex_list[iter]) );
	}

}

int create_mutexes (pthread_mutex_t* mutex_list, int cnt)
{
	int iter = 0;

	for (iter = 0; iter < cnt; iter++)
	{
		pthread_mutex_init ( &(mutex_list[iter]), NULL );
		pthread_mutex_lock ( &(mutex_list[iter]) );

		dprintf ("[%s][%02d] : [%8p]\n", 
				__FUNCTION__, iter,&(mutex_list[iter]) );
	}

	return iter;
}

int main (int argc, char** argv)
{
	int iter	= 0;
	state* states	= NULL;
	int num_tasks	= 0;

	pthread_mutex_t* mutex_list	= NULL;
	pthread_t* task_list		= NULL;

	pthread_attr_t	attr;
	struct sched_param param;

	if (argc < NUM_ARGS)
	{
		dprintf ("[%s] Usage: ./%s <num_of_tasks>\n", 
				argv[0], argv[0]);
		return (-1);
	}
	num_tasks = atoi (argv[1]);

	printf ("[%s] Ctxt_switch latency measurement, %d threads\n", 
			argv[0], num_tasks);

	task_list	= (pthread_t*) malloc (num_tasks * sizeof(pthread_t));
	states		= (state*) malloc (num_tasks * sizeof(state));
	mutex_list	= (pthread_mutex_t*) 
				malloc (num_tasks * sizeof(pthread_mutex_t));

	mlockall (MCL_CURRENT | MCL_FUTURE);

	/* Create message-queues */
	if ((create_mutexes(mutex_list, num_tasks)) < 0)
		return -1;

	/* Create tasks-list for dry run */ 
	pthread_attr_init (&attr);
	pthread_attr_setstacksize (&attr, TASK_STACK_SIZE);

	/* Create master-thread */
	states[0].idx		= iter;
	states[0].iter		= NUM_ITER;
	param.sched_priority	= TASK_PRIO ;
	states[0].rd_lock	= &(mutex_list[num_tasks - 1]);
	states[0].wr_lock	= &(mutex_list[0]);
	pthread_attr_setschedparam (&attr, &param);
	pthread_create ( &(task_list[iter]), 
			&attr, rt_task_master, (void*)&(states[0]));

	/* create slave threads */
	for (iter = 1; iter < num_tasks; iter++)
	{
		states[iter].idx	= iter;
		states[iter].iter	= NUM_ITER;
		states[iter].rd_lock	= &(mutex_list[iter - 1]);
		states[iter].wr_lock	= &(mutex_list[iter]);

		param.sched_priority	= TASK_PRIO - iter;
		pthread_attr_setschedparam (&attr, &param);
		pthread_create ( &(task_list[iter]), 
				&attr, rt_task_slave, (void*)&(states[iter]));
	}

	for (iter = 0; iter < num_tasks; iter++)
		pthread_join (task_list[iter], NULL);

	printf ("[%s] overhead: [%f], actual: [%f], result:[%f]\n", argv[0],
			(overhead), (actual), ( ((actual) / (NUM_ITER * num_tasks)) - 
						 (overhead / NUM_ITER)));

	destroy_mutexes(mutex_list, num_tasks);

	printf ("[%s] tasks terminated...\n", argv[0]);

	return 0;
}
